@@ -69,33 +69,22 @@ Note that exclude items are case-sensitive and must be matched exactly.
 
 
 parser.add_argument(
-    "-x", "--exclude",
-    metavar="PATH",
-    dest="exclude_items",
-    nargs="+",
-    help="add exclude item for parts of the configuration which will "
-         "be omitted from the comparison (these will be added to those "
-         "read from a file with the -y option, if both are used), "
-         "specified as a dictionary path, with keys separated by "
-         "colons")
-
-parser.add_argument(
-    "-y", "--exclude-filename",
+    "-r", "--rules-tree-filename",
     metavar="FILENAME",
-    help="read exclude items from lines in file")
+    help="read rules data from a YAML file (assumed to be an "
+         "'excludes' list by default - see -s option)")
 
 parser.add_argument(
-    "-z", "--exclude-yaml-filename",
-    metavar="FILENAME",
-    help="read exclude items from a YAML file")
-
-parser.add_argument(
-    "-Z", "--exclude-yaml-path",
+    "-s", "--rule-spec",
     metavar="PATH",
-    default="",
-    help="path into file specified with -z to where exclude items are "
-         "held, specified as a dictionary path, with keys separated by "
-         "colons")
+    action="append",
+    default=[],
+    help="add a rule to the rules list; rules are in the form '!path' "
+         "and processed in order: '!' excludes the specified items and "
+         "omitting it includes ONLY the specified items; the 'path' is "
+         "a dictionary path, with keys separated by colons; the path "
+         "is into the file read with -r; default can be viewed with "
+         "-R option)")
 
 parser.add_argument(
     "-e", "--explain",
@@ -115,15 +104,19 @@ parser.add_argument(
          "exclude and conversion process")
 
 parser.add_argument(
-    "-X", "--debug-excludes",
-    action="store_true",
-    help="dump the tree of items to be excluded from the comparions")
+    "-R", "--debug-rules",
+    action="count",
+    default=0,
+    help="dump the tree of items to be included and excluded from the "
+         "comparisons, based on the rules (multiple levels increase "
+         "verbosity up to a maximum of 2)")
 
 parser.add_argument(
     "-P", "--debug-parser",
-    action="store_true",
-    help="enable debugging of contextual parsing; assumed by default "
-         "if only a 'from' file is specified")
+    action="count",
+    default=0,
+    help="enable debugging of contextual parsing (multiple levels "
+         "increase verbosity up to a maxmimum of 2)")
 
 parser.add_argument(
     "-C", "--debug-config",
@@ -191,14 +184,12 @@ args = parser.parse_args()
 
 config_parser_class, config_diff_class = PLATFORMS[args.platform]
 
-exclude_items = args.exclude_items
-exclude_filename = args.exclude_filename
-exclude_yaml_filename = args.exclude_yaml_filename
-exclude_yaml_path = args.exclude_yaml_path
+rules_tree_filename = args.rules_tree_filename
+rule_specs = args.rule_spec
 explain = args.explain
 quiet = args.quiet
 no_output = args.no_output
-debug_excludes = args.debug_excludes
+debug_rules = args.debug_rules
 debug_parser = args.debug_parser
 debug_config = args.debug_config
 debug_diff = args.debug_diff
@@ -209,6 +200,19 @@ output_filename = args.output_filename
 devicenames = args.devicenames
 
 
+# if the rule specifications were not specified, assume this set by
+# default (we can't add_argument(default=...) option because the
+# 'action="append"' option just adds user-specified items to it)
+
+if not rule_specs:
+    rule_specs = [
+        "!rules:exclude:ALL",
+        "'!rules:exclude:%",
+        "rules:include:ALL",
+        "rules:include:%",
+    ]
+
+
 # check a couple of nonsensical configurations aren't being use related
 # to multiple devices
 
@@ -267,42 +271,23 @@ elif len(devicenames) > 1:
 diff = config_diff_class(explain, debug_config, debug_diff, debug_convert)
 
 
-# add any excludes specified on the command line
+# if a rules tree file was specified, read that in
 
-if exclude_items:
-    # add the exclude items specified directly on the command line (do
-    # this before reading those from a file, if specified, below)
-
-    diff.add_excludes(exclude_items)
+if rules_tree_filename:
+    diff.add_rules_tree_file(rules_tree_filename)
 
 
-# if an exclude [simple] file was specified, read and add those
+# dump the rule tree, if debugging enabled for it (this will include any
+# rules included by default, in the class)
 
-if exclude_filename:
-    diff.read_excludes(exclude_filename)
+if debug_rules >= 1:
+    print(">> rule tree:")
+    print(yaml.dump(diff.get_rules_tree(), default_flow_style=False))
+    print(">> rule specs:")
+    for spec in rule_specs:
+        print("->", spec)
 
-
-# if an exclude YAML file was specified, read that in
-
-if exclude_yaml_filename:
-    try:
-        diff.read_excludes_yaml(exclude_yaml_filename, exclude_yaml_path)
-
-    except TypeError:
-        print("error: exclude YAML path references non-dictionary item "
-              "in file", file=sys.stderr)
-
-        exit(1)
-
-    except KeyError:
-        print("error: exclude YAML path not found in file", file=sys.stderr)
-
-        exit(1)
-
-
-if debug_excludes:
-    print(">> exclude items:", yaml.dump(diff.get_excludes(),
-          default_flow_style=False), sep="\n", file=sys.stderr)
+    print()
 
 
 
@@ -355,8 +340,7 @@ def diffconfig(from_filename, to_filename, output_filename=None,
 
     if output_filename:
         if debug_convert:
-            print("debug: writing to output file: %s" % output_filename,
-                  file=sys.stderr)
+            print("debug: writing to output file:", output_filename)
 
         with open(output_filename, "w") as output_file:
             if diffs:
@@ -364,7 +348,7 @@ def diffconfig(from_filename, to_filename, output_filename=None,
 
     else:
         if debug_convert:
-            print("debug: writing to standard output", file=sys.stderr)
+            print("debug: writing to standard output")
 
         if diffs:
             print(diffs)
@@ -385,6 +369,26 @@ if devicenames:
         if not quiet:
             print(devicename)
 
+
+        # reset the rules list for this conversion - this is the same
+        # list each time, except the devicename changes and may cause
+        # some of the rules to change
+
+        diff.init_rules_active()
+        diff.add_rules_active(rule_specs, devicename)
+
+        if debug_rules >= 2:
+            num = 0
+            for spec, tree in diff.get_rules():
+                print(">> rule num: %d spec: %s" % (num, spec))
+                print(yaml.dump(tree, default_flow_style=False)
+                          if tree else "<none>\n", sep="\n")
+
+                num += 1
+
+
+        # do the actual conversion
+
         complete_success = diffconfig(
                                from_filename.replace("%", devicename),
                                to_filename.replace("%", devicename),
