@@ -10,7 +10,6 @@
 
 from copy import deepcopy
 import re
-import sys
 
 from deepops import (
     deepdiff, deepfilter, deepget, deepmerge, deepremoveitems, deepsetdefault)
@@ -74,7 +73,7 @@ class DiffConvert:
     wildcard match for all keys - for example, '("interface", None)' to
     match all interfaces.  The part of the configuration dictionary
     that matches the 'context + cmd' parts of the path will be passed
-    as the 'cfg' and 'diff' parameters to the action methods.
+    as the 'old/new' and 'rem/upd' parameters to the action methods.
 
     The child classes can override the following values:
 
@@ -86,14 +85,14 @@ class DiffConvert:
 
     cmd -- the path for this command in the configuration, following
     the context; this part of the dictionary will be passed in the
-    'cfg' and 'diff' parameters to the action methods; this variable
-    defaults to None and must be defined by a child class
+    'old/new' and 'rem/upddiff' parameters to the action methods; this
+    variable defaults to None and must be defined by a child class
 
     ext -- an extension to path which must be matched, but this part is
     not used to further specify the part of the dictionaries passed to
-    the action functions for the 'cfg' and 'diff' parameters: this is
-    useful when a higher level in the configuration dictionary is
-    required in the action function
+    the action functions for the 'old/new' and 'rem/upd' parameters:
+    this is useful when a higher level in the configuration dictionary
+    is required in the action function
 
     name -- if there is more than one action acting at a particular path
     point, this can be defined to differentiate them: it's useful when
@@ -241,7 +240,7 @@ class DiffConvert:
             return False
 
 
-    def remove(self, cfg, diff, args):
+    def remove(self, old, rem, args):
         """The remove() method is called when the specified path is in
         the remove differences (i.e. it's in the 'from' configuration
         but not in the 'to' configuration), unless the containing
@@ -249,11 +248,11 @@ class DiffConvert:
 
         Keyword arguments:
 
-        cfg -- the value of the dictionary item at the matching path in
+        old -- the value of the dictionary item at the matching path in
         the 'from' configuration dictionary (sometimes, the full details
         of the old configuration may be required to remove it)
 
-        diff -- the value of the dictionary item at the matching path in
+        rem -- the value of the dictionary item at the matching path in
         the remove differences dictionary (sometimes, it's necessary to
         know only what is removed from the new configuration)
 
@@ -273,7 +272,7 @@ class DiffConvert:
         pass
 
 
-    def update(self, cfg, diff, args):
+    def update(self, new, upd, args):
         """The update() method is called when the specified path is in
         the update differences (i.e. it's in the 'to' configuration but
         not in the 'from' configuration, or the value differs between
@@ -281,11 +280,11 @@ class DiffConvert:
 
         Keyword arguments:
 
-        cfg -- the value of the dictionary item at the matching path in
+        new -- the value of the dictionary item at the matching path in
         the 'to' configuration dictionary (sometimes, teh full details
         of the new configuration may be required to update it)
 
-        diff -- the value of the dictionary item at the matching path in
+        upd -- the value of the dictionary item at the matching path in
         the update differences dictionary
 
         args -- [only] the arguments in the path which were matched with
@@ -340,7 +339,7 @@ class DiffConvert:
     def get_ext(self, cfg):
         """This method gets the extension part of the path, given a
         configuration dictionary starting at the path (i.e. what is
-        passed as 'cfg' and 'diff' in the action methods).
+        passed as 'old/new' and 'rem/upd' in the action methods).
 
         An action method [remove() or update()] can use this to get the
         extension portion without needing to explicitly index through
@@ -385,14 +384,14 @@ class DiffConfig:
         by the DiffConvert objects (if available).
 
         init_debug_config=False -- dump the 'from' and 'to'
-        configurations (after exludes) to stderr
+        configurations (after exludes)
 
         init_debug_diff=False -- dump the differences (remove and update
-        configurations) to stderr
+        configurations)
 
         init_debug_convert=0 -- level of debugging information for the
-        conversion process: >= 1: include steps, >= 2: include 'cfg' and
-        'diff' parameters
+        conversion process: >= 1: include steps, >= 2: include 'old/new'
+        and 'rem/upd' parameters
         """
 
         # store the [initial] settings for the conversion process
@@ -403,10 +402,10 @@ class DiffConfig:
 
         # initialise the dictionary of excludes, which will be passed
         # to deepremoveitems()
-        self.init_excludes()
+        self.init_rules()
 
         # initialise the list of converters and add them
-        self._cvtrs = []
+        self._cvts = []
         self._add_converters()
 
 
@@ -422,7 +421,7 @@ class DiffConfig:
         pass
 
 
-    def _add_converter(self, cvtr):
+    def _add_converter(self, cvt):
         """Add an individual convert object, a child of the
         DiffConverter class, to the list of converters to be used by the
         convert() method.
@@ -433,26 +432,26 @@ class DiffConfig:
         be found, a KeyError will be raised.
         """
 
-        if not cvtr.insert_before:
+        if not cvt.insert_before:
             # no 'insert_before' - just add it to the end
-            self._cvtrs.append(cvtr)
+            self._cvts.append(cvt)
 
         else:
             # 'insert_before' so we need to find that entry
 
             # get the indexes and entries in the converter list
-            for pos, chk_cvtr in enumerate(self._cvtrs):
+            for pos, chk_cvt in enumerate(self._cvts):
                 # if this entry is the one we're inserting before, put
                 # it here and stop
-                if chk_cvtrs.name == cvtrs.insert_before:
-                    self._cvtrs.insert(pos, cvtr)
+                if chk_cvts.name == cvts.insert_before:
+                    self._cvts.insert(pos, cvt)
                     break
 
             else:
                 # we got the end of the list and didn't find it, so
                 # raise an exception
                 raise KeyError("cannot find existing converter with name: "
-                               + cvtr.insert_before)
+                               + cvt.insert_before)
 
 
     def _explain_comment(self, path):
@@ -500,8 +499,8 @@ class DiffConfig:
         return []
 
 
-    def init_excludes(self):
-        """This method initialises the excludes dictionary.  In the base
+    def init_rules(self):
+        """This method initialises the rules s.  In the base
         class, it the resets it to an empty dictionary, but child
         classes can extend this to add some default exclusions for
         standard system configuration entries which should not be
@@ -512,92 +511,145 @@ class DiffConfig:
         required).
         """
 
-        self._excludes = {}
+        self.init_rules_tree()
+        self.init_rules_active()
 
 
-    def add_exclude(self, path):
-        """Add a single exclude item to the dictionary.  The exclude is
-        supplied as a string, with the levels in the path of the
-        configuration dictionary separated by colons (":"), and then
-        added to the excludes dictionary.
+    def init_rules_tree(self):
+        """This method initialises the rules tree (the dictionary of
+        rules typically read from a file.
+
+        In the base class, it just sets it to an empty dictionary but
+        some platform-specific classes may wish to extend this to set up
+        a default tree (along with init_rules_active()).
         """
 
-        # start at the top of the excludes dictionary
-        excludes_sub = self._excludes
-
-        # work through the levels in the exclude item path
-        for level in path.split(":"):
-            # create this level of the excludes dictionary as a new
-            # empty dictionary and move down to it for the next level
-            excludes_sub = excludes_sub.setdefault(level, {})
+        self._rules_tree = {}
 
 
-    def add_excludes(self, paths):
-        """Add a list of exclude items as strings to the excludes, using
-        add_exclude().
+    def init_rules_active(self):
+        """This method initialises the active rules list (the list of
+        rules specifying what should be used from the rules tree).
+
+        In the base class, it just sets it to an empty list but some
+        platform-specific classes may wish to extend this to set up a
+        default list.
         """
 
-        for path in paths:
-            self.add_exclude(path)
+        self._rules_active = []
 
 
-    def read_excludes(self, filename):
-        """Read a list of exclude items from a file and add them to the
-        excludes dictionary using add_exclude().
+    def add_rules_tree_file(self, filename):
+        """Read a tree of rules items from a YAML file.  This is
+        typically done once but then different portions of the rules
+        dictionary selected with set_rules_active().
 
-        A hash ('#') and anything following it is stripped out as a
-        comment.  Trailing whitespace is also removed.  Blank lines
-        (including those created by the above actions) are skipped.
+        The contents of the file are added to the current tree.  To
+        clear the current tree first, use init_rules_tree().
         """
 
-        # work through the lines in the file
-        for l in open(filename):
-            # strip out anything following a '#' (indicating comment)
-            # and any trailing whitespace
-            l = re.sub("#.*", "", l)
-            l = l.rstrip()
-
-            # if the line is not blank, add it as an exclude item
-            if l:
-                self.add_exclude(l)
-
-
-    def read_excludes_yaml(self, filename, path=None):
-        """Read a list of exclude items from a YAML file and add them to
-        the excludes dictionary.
-
-        The path parameter specifies the key in the dictionary where the
-        excludes are specified.  This is useful if the dictionary is
-        also contains other data (such as the inventory).  The path is
-        specified a list of keys, separated by colons.  If the path is
-        None or otherwise False, the top of the dictionary is used.
-        """
-
-
-        # try to read in the YAML excludes file
-
         try:
-            f = yaml.safe_load(open(filename))
+            file_rules_tree = yaml.safe_load(open(filename))
 
         except yaml.parser.ParserError as exception:
-            raise ValueError("failed parsing YAML excludes file: %s: %s"
+            raise ValueError("failed parsing rules file: %s: %s"
                                  % (filename, exception))
 
-
-        # if no path was specified, just use the entire file, else get
-        # the key given
-
-        if not path:
-            deepmerge(self._excludes, f)
-        else:
-            deepmerge(self._excludes,
-                      deepget(f, *path.split(":"), default_error=True))
+        self._rules_tree.update(file_rules_tree)
 
 
-    def get_excludes(self):
-        "This method just returns the excludes dictionary."
+    def add_rules_active(self, rule_specs, devicename):
+        """Add a list of rules to the current rule list, which specifies
+        what parts of the rules tree should be used and how (include or
+        exclude these items), to control the comparison done by the
+        convert() method.
 
-        return self._excludes
+        The rules are specified as a list of strings in the format
+        '[!]<path>' where '!' means 'exclude' (if omitted, it means
+        'include') and 'path' is a colon-separated list of keys giving
+        the path into the rules tree.  A particular element can be
+        given as '%', in which case the 'devicename' parameter will be
+        used.
+
+        For example, if '!device-excludes:%' is given, and the
+        devicename is 'router1', the part of the rules tree indexed by
+        ["device-excludes"]["router1"] will be excluded from the
+        comparison.
+
+        The rules given will be added to the current rules list; if the
+        list is to be cleared first, use init_rules_list().
+        """
+
+
+        for rule_spec in rule_specs:
+            include = not rule_spec.startswith("!")
+            if not include:
+                rule_spec = rule_spec[1:]
+
+            path = [ devicename if i == "%" else i
+                                  for i in rule_spec.split(":") ]
+
+            self._rules_active.append( (include, path) )
+
+
+    def get_rules_tree(self):
+        """The method returns the current rules tree (as initialised by
+        init_rules_tree() and extended with read_rules_tree()).  This
+        should not be modified.
+        """
+
+        return self._rules_tree
+
+
+    def get_rules(self):
+        """This method just returns the active rules list and the
+        portion of the rules tree it applies to.
+
+        The return value is a list, one entry per rule, containing
+        2-tuples: the first entry is the rule specification (a string,
+        in the format described by add_rules_active()), and the second
+        the portion of the rules tree that the path references (or None,
+        if that part of the tree does not exist).
+
+        This method is mainly used for debugging messages.
+        """
+
+        r = []
+        for include, path in self._rules_active:
+            r.append(
+                (("" if include else "!") + ":".join(path),
+                 deepget(self._rules_tree, *path) ) )
+
+        return r
+
+
+    def apply_rules(self, d):
+        """This method applies the current rules (tree and active list)
+        to the supplied configuration dictionary, either only including
+        what's specified (using deepfilter()) or excluding (using
+        deepremoveitems()) the specified portions.
+
+        The configuration dictionary is modified in place.
+
+        The method can be used on a configuration dictionary, or the
+        remove/update dictionaries returned by deepdiff().
+        """
+
+        for include, path in self._rules_active:
+            # get the porttion of the rules tree specified by the path
+            # in this rule
+            path_dict = deepget(self._rules_tree, *path)
+
+            # skip to the next entry, if this part was not found
+            if path_dict is None:
+                continue
+
+            # do either the include or exclude on the configuration
+            # dictionary, in place)
+            if include:
+                d = deepfilter(d, path_dict)
+            else:
+                deepremoveitems(d, path_dict)
 
 
     def convert(self, from_cfg, to_cfg):
@@ -615,32 +667,30 @@ class DiffConfig:
         """
 
 
-        # remove any excludes from the 'from' config
-
-        deepremoveitems(from_cfg, self._excludes)
+        self.apply_rules(from_cfg)
 
         if self._debug_config:
-            print(">> 'from' configuration:",
-                yaml.dump(from_cfg, default_flow_style=False), sep="\n",
-                file=sys.stderr)
+            print(">> 'from' configuration (after rules, if specified):",
+                  yaml.dump(dict(from_cfg), default_flow_style=False),
+                  sep="\n")
 
 
         # if no 'to' config was specified, stop here (we assume we're
         # just testing, parsing and excluding items from the 'from'
         # configuration and stopping
+        #
+        # we check for None explicitly for the difference between no
+        # configuration and an empty configuration
 
-        if not to_cfg:
+        if to_cfg is None:
             return None
 
 
-        # remove any excludes from the 'to' config
-
-        deepremoveitems(to_cfg, self._excludes)
+        self.apply_rules(to_cfg)
 
         if self._debug_config:
-            print(">> 'to' configuration:",
-                yaml.dump(to_cfg, default_flow_style=False), sep="\n",
-                file=sys.stderr)
+            print(">> 'to' configuration (after rules, if specified):",
+                yaml.dump(dict(to_cfg), default_flow_style=False), sep="\n")
 
 
         # use deepdiff() to work out the differences between the two
@@ -665,7 +715,7 @@ class DiffConfig:
                   "=> differences - update (add/change):",
                   yaml.dump(update_cfg, default_flow_style=False),
 
-                  sep="\n", file=sys.stderr)
+                  sep="\n")
 
 
         # initialise the list of changes (the returned configuration
@@ -676,12 +726,12 @@ class DiffConfig:
 
         # go through the list of converter objects
 
-        for cvtr in self._cvtrs:
+        for cvt in self._cvts:
             if self._debug_convert:
                 print(">> checking difference:",
                       pathstr([ "*" if i is None else i
-                                    for i in cvtr.get_path() ]),
-                      "name:", cvtr.name, file=sys.stderr)
+                                    for i in cvt.get_path() ]),
+                      "name:", cvt.name)
 
 
             # get all the remove and update matches for this converter
@@ -694,8 +744,8 @@ class DiffConfig:
             # we combine them by manually building a list, rather than
             # using a set as a set requires hashable types
 
-            remove_matches = cvtr.full_matches(remove_cfg_full)
-            update_matches = cvtr.full_matches(update_cfg)
+            remove_matches = cvt.full_matches(remove_cfg_full)
+            update_matches = cvt.full_matches(update_cfg)
 
             all_matches = []
             for match in sorted(remove_matches + update_matches):
@@ -709,26 +759,24 @@ class DiffConfig:
                 if match in remove_matches:
                     if self._debug_convert:
                         print("=> remove match:",
-                              pathstr(match, cvtr.wildcard_indices),
-                              file=sys.stderr)
+                              pathstr(match, cvt.wildcard_indices))
 
 
                     # if we're removing the entire context for this
                     # match, we don't need to do it, as that will
                     # remove this, while it's at it
 
-                    if cvtr.context_removed(remove_cfg, match):
+                    if cvt.context_removed(remove_cfg, match):
                         if self._debug_convert:
                             print(
-                                "-> containing context being removed - skip",
-                                file=sys.stderr)
+                                "-> containing context being removed - skip")
 
                         continue
 
 
                     # get elements in the path matching wildcards
 
-                    args = cvtr.get_wildcard_args(match)
+                    args = cvt.get_wildcard_args(match)
 
 
                     # get the 'from' and 'remove' parts of the
@@ -736,22 +784,22 @@ class DiffConfig:
                     # for the path specified in the converter (ignoring
                     # the extension 'ext')
 
-                    cvtr_cfg = cvtr.get_cfg(from_cfg, match)
-                    cvtr_diff = cvtr.get_cfg(remove_cfg, match)
+                    cvt_old = cvt.get_cfg(from_cfg, match)
+                    cvt_rem = cvt.get_cfg(remove_cfg, match)
 
                     if self._debug_convert >= 2:
                         print(
                             "-> from configuration:",
-                            yaml.dump(cvtr_cfg, default_flow_style=False),
+                            yaml.dump(cvt_old, default_flow_style=False),
                             "-> remove configuration:",
-                            yaml.dump(cvtr_diff, default_flow_style=False),
-                            sep="\n", file=sys.stderr)
+                            yaml.dump(cvt_rem, default_flow_style=False),
+                            sep="\n")
 
 
                     # call the remove converter action method using the
                     # corresponding entry from the 'from' configuration
 
-                    change = cvtr.remove(cvtr_cfg, cvtr_diff, args)
+                    change = cvt.remove(cvt_old, cvt_rem, args)
 
 
                     # if some changes were returned by the action, add
@@ -766,7 +814,7 @@ class DiffConfig:
                             change = [change]
 
                         if self._debug_convert:
-                            print("\n".join(change), file=sys.stderr)
+                            print("\n".join(change))
 
 
                         # add a comment, explaining the match, if
@@ -786,7 +834,7 @@ class DiffConfig:
 
                     else:
                         if self._debug_convert:
-                            print("-> no action", file=sys.stderr)
+                            print("-> no action")
 
 
                 # handle UPDATE conversions, if matching
@@ -794,11 +842,10 @@ class DiffConfig:
                 if match in update_matches:
                     if self._debug_convert:
                         print("=> update match:",
-                              pathstr(match, cvtr.wildcard_indices),
-                              file=sys.stderr)
+                              pathstr(match, cvt.wildcard_indices))
 
 
-                    args = cvtr.get_wildcard_args(match)
+                    args = cvt.get_wildcard_args(match)
 
 
                     # get the 'from' and 'update' parts of the
@@ -806,21 +853,21 @@ class DiffConfig:
                     # for the path specified in the converter (ignoring
                     # the extension 'ext')
 
-                    cvtr_cfg = cvtr.get_cfg(to_cfg, match)
-                    cvtr_diff = cvtr.get_cfg(update_cfg, match)
+                    cvt_new = cvt.get_cfg(to_cfg, match)
+                    cvt_upd = cvt.get_cfg(update_cfg, match)
 
                     if self._debug_convert >= 2:
                         print(
                             "-> to configuration:",
-                            yaml.dump(cvtr_cfg, default_flow_style=False),
+                            yaml.dump(cvt_new, default_flow_style=False),
                             "-> update configuration:",
-                            yaml.dump(cvtr_diff, default_flow_style=False),
-                            sep="\n", file=sys.stderr)
+                            yaml.dump(cvt_upd, default_flow_style=False),
+                            sep="\n")
 
 
                     # call the update converter action method
 
-                    change = cvtr.update(cvtr_cfg, cvtr_diff, args)
+                    change = cvt.update(cvt_new, cvt_upd, args)
 
 
                     if change is not None:
@@ -828,7 +875,7 @@ class DiffConfig:
                             change = [change]
 
                         if self._debug_convert:
-                            print("\n".join(change), file=sys.stderr)
+                            print("\n".join(change))
 
 
                         if self._explain:
@@ -842,14 +889,14 @@ class DiffConfig:
 
                     else:
                         if self._debug_convert:
-                            print("-> no action", file=sys.stderr)
+                            print("-> no action")
 
 
             # print a blank line if debugging, just to space things out
             # nicely
 
             if self._debug_convert:
-                print(file=sys.stderr)
+                print()
 
         # if nothing was generated, just return nothing
 
