@@ -17,7 +17,7 @@ This module parses Cisco IOS configuration files into a dictionary.
 
 import re
 
-from deepops import deepsetdefault
+from deepops import deepsetdefault, deepget
 from netaddr import IPNetwork
 
 from .configparser import IndentedContextualCommand, IndentedContextualConfig
@@ -586,13 +586,25 @@ class _Cmd_Int(_Cmd):
         #
         # TODO: we really should handle this more flexibly, with a
         # central funcation matching regexps of the interface name
-        match = re.match(r"^([-a-z]+)([0-9/.]*)$", name.lower())
+        match = re.match(r"([-a-z]+)([0-9/.]*)$", name.lower())
         if match:
             _type, num = match.groups()
             _type = _INTERFACE_CANONICALS.get(_type, _type).capitalize()
             name = _type + num
 
-        return deepsetdefault(cfg, "interface", name)
+        i = deepsetdefault(cfg, "interface", name)
+
+        # IOS has an odd behaviour that, when an interface is created in
+        # configure mode, it will default to shutdown or not, depending
+        # on its type; in startup configurations, however, they are
+        # always not shutdown
+        #
+        # we default to not shutdown, unless this has been explicitly
+        # overridden: this has the effect of 'no shutdown'ing the
+        # interface, if it is being created
+        i.setdefault("shutdown", False)
+
+        return i
 
 _cmds.append(_Cmd_Int)
 
@@ -725,6 +737,15 @@ class _Cmd_Int_IPPIMBSRBdr(_CmdContext_Int):
 _cmds.append(_Cmd_Int_IPPIMBSRBdr)
 
 
+class _Cmd_Int_IPProxyARP(_CmdContext_Int):
+    cmd = r"(?P<no>no )?ip proxy-arp"
+
+    def action(self, groups, cfg, params):
+        cfg["ip-proxy-arp"] = not bool(groups["no"])
+
+_cmds.append(_Cmd_Int_IPProxyARP)
+
+
 class _Cmd_Int_ServPol(_CmdContext_Int):
     cmd = r"service-policy (?P<policy>.+)"
 
@@ -738,12 +759,7 @@ class _Cmd_Int_Shutdown(_CmdContext_Int):
     cmd = r"(?P<no>no )?shutdown"
 
     def action(self, groups, cfg, params):
-        if groups["no"]:
-            # we allow interfaces to be 'no shutdown' so have to clear
-            # the shutdown option
-            cfg.pop("shutdown", None)
-        else:
-            cfg["shutdown"] = None
+        cfg["shutdown"] = not groups["no"]
 
 _cmds.append(_Cmd_Int_Shutdown)
 
@@ -769,6 +785,59 @@ class _Cmd_Int_StandbyIPSec(_CmdContext_Int):
 _cmds.append(_Cmd_Int_StandbyIPSec)
 
 
+class _Cmd_Int_StandbyIP6(_CmdContext_Int):
+    cmd = r"standby (?P<grp>\d+) ipv6 (?P<addr>\S+)"
+
+    def action(self, groups, cfg, params):
+        deepsetdefault(
+            cfg, "standby", "group", int(groups["grp"]), "ipv6", last=set()
+        ).add(groups["addr"])
+
+_cmds.append(_Cmd_Int_StandbyIP6)
+
+
+class _Cmd_Int_StandbyPreempt(_CmdContext_Int):
+    cmd = r"standby (?P<grp>\d+) preempt"
+
+    def action(self, groups, cfg, params):
+        deepsetdefault(
+            cfg, "standby", "group", int(groups["grp"]))["preempt"] = None
+
+_cmds.append(_Cmd_Int_StandbyPreempt)
+
+
+class _Cmd_Int_StandbyPri(_CmdContext_Int):
+    cmd = r"standby (?P<grp>\d+) priority (?P<pri>\d+)"
+
+    def action(self, groups, cfg, params):
+        deepsetdefault(
+            cfg, "standby", "group", int(groups["grp"]))["priority"] = (
+            int(groups["pri"]))
+
+_cmds.append(_Cmd_Int_StandbyPri)
+
+
+class _Cmd_Int_StandbyTimers(_CmdContext_Int):
+    cmd = r"standby (?P<grp>\d+) timers (?P<timers>\d+ \d+)"
+
+    def action(self, groups, cfg, params):
+        deepsetdefault(
+            cfg, "standby", "group", int(groups["grp"]))["timers"] = (
+            groups["timers"])
+
+_cmds.append(_Cmd_Int_StandbyTimers)
+
+
+class _Cmd_Int_StandbyTrk(_CmdContext_Int):
+    cmd = r"standby (?P<grp>\d+) track (?P<obj>\d+)( (?P<extra>.+))?"
+
+    def action(self, groups, cfg, params):
+        deepsetdefault(cfg, "standby", "group", int(groups["grp"]), "track"
+            )[groups["obj"]] = groups["extra"]
+
+_cmds.append(_Cmd_Int_StandbyTrk)
+
+
 class _Cmd_Int_StandbyVer(_CmdContext_Int):
     cmd = r"standby version (?P<ver>\d)"
 
@@ -984,16 +1053,104 @@ class _Cmd_STPPriority(_Cmd):
     cmd = r"spanning-tree vlan (?P<tags>[-0-9,]+) priority (?P<priority>\d+)"
 
     def action(self, groups, cfg, params):
-        priority = groups["priority"]
-
         cfg_stp_pri = cfg.setdefault("spanning-tree-vlan-priority", {})
         for tag in _expand_set(groups["tags"]):
-            cfg_stp_pri[int(tag)] = int(priority)
+            cfg_stp_pri[int(tag)] = int(groups["priority"])
 
 _cmds.append(_Cmd_STPPriority)
 
 
 
+# TRACK ...
+
+
+
+class _Cmd_TrackRoute(_Cmd):
+    cmd = (r"track (?P<obj>\d+)"
+           r" (?P<proto>ip|ipv6) route"
+           r" (?P<net>[0-9a-fA-F.:]+/\d+|[0-9.]+ [0-9.]+)"
+           r" (?P<extra>metric .+|reachability)")
+    enter_context = "track"
+
+    def action(self, groups, cfg, params):
+        # the 'net' can be in 'network netmask' or CIDR format, but the
+        # netaddr.IPNetwork() object requires a slash between the
+        # network and netmask, so we just change the space to a slash
+        net = IPNetwork(groups["net"].replace(" ", "/"))
+
+        # reconstruct a normalised version of the criterion
+        criterion = ("%s route %s %s"
+                        % (groups["proto"], net, groups["extra"]))
+
+        # create the new track object and store the criterion in it
+        t = deepsetdefault(cfg, "track", int(groups["obj"]))
+        t["criterion"] = criterion
+
+        # return the track object for the new context
+        return t
+
+_cmds.append(_Cmd_TrackRoute)
+
+
+class _Cmd_TrackOther(_Cmd):
+    cmd = r"track (?P<obj>\d+) (?P<other>(interface .+|list .+|stub-object))"
+    enter_context = "track"
+
+    def action(self, groups, cfg, params):
+        t = deepsetdefault(cfg, "track", int(groups["obj"]))
+        t["criterion"] = groups["other"]
+
+        return t
+
+_cmds.append(_Cmd_TrackOther)
+
+
+class _Cmd_Track(_Cmd):
+    cmd = r"track (?P<obj>\d+)"
+    enter_context = "track"
+
+    def action(self, groups, cfg, params):
+        # if there is no criterion, we're modifying an existing object,
+        # which must have already been defined, so we deliberately don't
+        # create it with deepsetdefault() but just deepget() it with
+        # default_error set, to force an error here, if it doesn't exist
+        return deepget(cfg, "track", int(groups["obj"]), default_error=True)
+
+_cmds.append(_Cmd_Track)
+
+
+class _CmdContext_Track(_Cmd):
+    context = "track"
+
+
+class _Cmd_Track_Delay(_CmdContext_Track):
+    cmd = r"delay (?P<delay>.+)"
+
+    def action(self, groups, cfg, params):
+        cfg["delay"] = groups["delay"]
+
+_cmds.append(_Cmd_Track_Delay)
+
+
+class _Cmd_Track_IPVRF(_CmdContext_Track):
+    cmd = r"ip vrf (?P<vrf>\S+)"
+
+    def action(self, groups, cfg, params):
+        cfg["ip-vrf"] = groups["vrf"]
+
+_cmds.append(_Cmd_Track_IPVRF)
+
+
+class _Cmd_Track_Obj(_CmdContext_Track):
+    cmd = r"object (?P<obj>.+)"
+
+    def action(self, groups, cfg, params):
+        deepsetdefault(cfg, "object", last=set()).add(groups["obj"])
+
+_cmds.append(_Cmd_Track_Obj)
+
+
+
 # VLAN ...
 
 
